"""
The database interface is helpful to interact with larger numbers
of PF objects in the database (getting, setting, storing or comparing
attributes).
One use case is for example to define a set of standard parameters for a
larger model. Using the interface, one can easily get, store (e.g. in a json file) and later reset a set of parameters.

ToDo: Add tutorial for this interface (when there is more functionality).
"""

import sys
from fnmatch import fnmatchcase
from typing import Iterable
import importlib, inspect

from powfacpy.applications.application_base import ApplicationBase
from powfacpy.base.string_manipulation import PFStringManipulation
from powfacpy.pf_classes.protocols import (
    PFGeneral,
    PFApp,
)


class Database(ApplicationBase):

    def __init__(
        self, pf_app: PFApp | None | bool = False, cached: bool = False
    ) -> None:
        super().__init__(pf_app, cached)

    def get_object_attributes(
        self,
        objs: Iterable[PFGeneral],
        class_attributes: dict = None,
        truncated_path: str = "",
        pf_obj_handling: str = "path",
    ) -> dict[str, dict[str, str]]:
        """
        Get dictionary with attributes of objects.
        The keys of the dictionary are the paths of the objects.
        The values are dictionaries with keys (attribute names) and values (attribute values).

        Args:
            objs(Iterable[PFGeneral]): Iterable with PF objects.

            class_attributes(dict, optional): dictionary with keys (class names) and values (iterable with attribute names). To control which attributes are relevant for a class. The class names can contain wildcards ("*"). Defaults to None.
            Example:
            ```python
            {
                "ElmTr2": ["typ_id"],
                "*": ["loc_name", ],
                "ElmVac": ["bus1", "outserv"],
            }
            ```
            -> "loc_name" is relevant for every class

            truncated_path(str, optional): If specified, this path is truncated from the keys (object paths). Defaults to "".
            Example:
            ```
            truncated_path = "Network Model\\Network Data"
            # then 'Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\test_database_interface\\Grid.ElmNet\\AC Voltage Source.ElmVac'
            # becomes (first part truncated):
            'test_database_interface\\Grid.ElmNet\\AC Voltage Source.ElmVac"
            ```

            pf_obj_handling (str, optional): If the value of an attribute is a PF object, there are several options on how to read the attribute (please see _handle_attribute_type_for_reading). Defaults to "path".

        Returns:
            dict[str, dict[str, str]]: PF objects' data
            Example:
            ```
            {
                "Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\test_database_interface\\Grid.ElmNet\\AC Voltage Source.ElmVac": {
                        "loc_name": "AC Voltage Source",
                        "bus1": "Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\test_database_interface\\Grid.ElmNet\\Terminal HV 1.ElmTerm\\Cub_1.StaCubic",
                        "outserv": 0
                },
                "Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\test_database_interface\\Grid.ElmNet\\Terminal HV 1.ElmTerm\\Cub_1.StaCubic": {
                        "loc_name": "Cub_1"
                },
            }
            ```
        """
        if not class_attributes:
            class_attributes = {
                "*": []
            }  # no relevant attributes -> get only keys (object paths)
        if truncated_path:  # include class names in truncated_path
            truncated_path = self.act_prj.get_path_of_obj_with_class_names(
                truncated_path
            )

        obj_attr_dict = {}
        for obj in objs:
            obj_path = self.act_prj.get_path_of_obj_with_class_names(obj)
            if truncated_path:
                obj_path = PFStringManipulation.truncate_until(
                    obj_path, truncated_path + "\\"
                )
            obj_attr_dict[obj_path] = {}
            for class_name in class_attributes.keys():
                if fnmatchcase(
                    obj.GetClassName(), class_name
                ):  # fnmatch allows to use wildcards ("*")
                    for attr in class_attributes[class_name]:
                        obj_attr_dict[obj_path][attr] = (
                            self._handle_attribute_type_for_reading(
                                obj, attr, pf_obj_handling
                            )
                        )
        return obj_attr_dict

    def set_object_attributes(self, obj_attr_dict: dict, added_path: str = "") -> None:
        """
        Takes a dict with keys (object paths) and values (dict with attributes
        and their values) and writes the data to the PF database.
        (the dict can be created e.g. with 'get_object_attributes')

        Arguments:
          obj_attr_dict (dict): data

          added_path (str): Assumes that the object paths are relative to a parent folder inside the project. Adds added_path to the paths.
        """
        for obj, attr_key_val in obj_attr_dict.items():
            if added_path:
                obj = added_path + "\\" + obj
            obj = self.act_prj.get_unique_obj(obj, include_subfolders=False)
            for attr, value in attr_key_val.items():
                if isinstance(value, str):
                    if not isinstance(
                        obj.GetAttributeType(attr), str
                    ):  # Then a PF object is expected
                        value_pf_obj = self.act_prj.get_unique_obj(
                            value, error_if_non_existent=False, include_subfolders=False
                        )
                        if value_pf_obj:
                            value = value_pf_obj
                self.act_prj.set_attr(obj, {attr: value})

    def _handle_attribute_type_for_reading(
        self, obj: PFGeneral, attr: str, pf_obj_handling="path", truncated_path=""
    ) -> PFGeneral | str:
        """
        Various ways to handle attributes which are PF objects (i.e. a PF object holds another PF object as an attribute, e.g. a 'type') when
        reading from the PF database.

        PF objects can be read as
          - a PF object (pf_obj_handling="original_pf_obj")
          - the path of the object (pf_obj_handling="path")

        Arguments:
            obj (PFGeneral): PF object
            attr (str): Attribute of the object
            pf_obj_handling (str, optional): see above. Defaults to "path".
            truncated_path: If pf_obj_handling="path", then 'truncated_path' is truncated from the beginning of the object's path. Defaults to "".

        Returns:
            PFGeneral | str: PF object or its path
        """
        attr_value = self.act_prj.get_attr(obj, attr)
        if isinstance(attr_value, (str, int, float)) or not attr_value:
            return attr_value
        else:  # Then it must be a PF object
            if pf_obj_handling == "path":
                if truncated_path:
                    truncated_path = self.act_prj.get_path_of_obj_with_class_names(
                        truncated_path
                    )
                return self.act_prj.get_path_of_obj_with_class_names(attr_value)
            elif pf_obj_handling == "original_pf_obj":
                return attr_value

    def make_loc_name_of_calc_relevant_objects_unique(
        self, pf_classes: Iterable[str] | str = "Elm*"
    ) -> dict[str, int]:
        if not isinstance(pf_classes, Iterable):
            pf_classes = self.get_class_names(pf_classes)
        for class_name in pf_classes:
            objs: list[PFGeneral] = self.act_prj.app.GetCalcRelevantObjects(
                f"*.{class_name}"
            )
            loc_names_count = {}
            for obj in objs:
                if not obj.loc_name in loc_names_count.keys():
                    loc_names_count[obj.loc_name] = 0
                else:
                    loc_names_count[obj.loc_name] += 1
                    obj.loc_name = obj.loc_name + f"_{loc_names_count[obj.loc_name]}"
        return loc_names_count

    def get_class_names(self, name: str = "*") -> list[str]:
        class_names = []
        for cls_name, _cls in inspect.getmembers(
            importlib.import_module("powfacpy.pf_classes.protocols"), inspect.isclass
        ):
            if fnmatchcase(cls_name, name):
                class_names.append(cls_name)
        return class_names

    # Todo: define standard class attributes
    def get_standard_class_attributes():
        return {"ElmTerm": ["Unom"]}



from __future__ import annotations

import os
from os import replace

import pandas as pd
import numpy as np
from icecream import ic

from powfacpy.pf_classes.protocols import ElmFile
from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.base.active_project import ActiveProjectCached


class MeasurementFile(ElmBase):

    __slots__ = ()

    def __init__(self, obj: ElmFile) -> None:
        super().__init__(obj)
        self._obj: ElmFile

    def __new__(cls, *args, **kwargs) -> ElmFile | MeasurementFile:
        """Implemented only to add type hints for the created instance.

        Returns:
            ElmTerm | Terminal: New instance
        """
        instance = super().__new__(cls)
        return instance

    def from_pandas_dataframe(
        self,
        dataframe: pd.DataFrame,
        csv_file_name: str,
        csv_file_dir: str | None = None,
    ) -> str:
        """Read data from a pandas dataframe.

        A csv file is created from the dataframe and the measurement file reads from the csv file.

        Args:
            dataframe (pd.DataFrame): Data for measurement file. IMPORTANT: The index of the frame is the time.

            csv_file_name (str): Name of csv file (created using the dataframe) from which the measurement file reads.

            csv_file_dir (str | None, optional): directory where csv file is created. Must have write access. Defaults to None (external data directory of PF is used, see Settings\Project Settings).

        Returns:
            str: Directory where csv file was created
        """
        act_prj = ActiveProjectCached()
        if not csv_file_dir:
            csv_file_dir_with_special_chars = (
                "$(ExtDataDir)\\"
                + self.get_path_of_elmfile_data_inside_external_data_directory()
            )
            csv_file_dir = act_prj._replace_special_PF_characters_in_path_string(
                csv_file_dir_with_special_chars
            )
            if not os.path.exists(csv_file_dir):
                os.makedirs(csv_file_dir)
        else:
            csv_file_dir_with_special_chars = csv_file_dir

        np_array = np.concatenate(
            [dataframe.index.to_numpy().reshape(-1, 1), dataframe.to_numpy()], axis=1
        )
        first_row = np.zeros((1, np_array.shape[1]))
        first_row[0, 0] = np_array.shape[1]
        np_array = np.concatenate([first_row, np_array])
        np.savetxt(
            csv_file_dir + "\\" + csv_file_name + ".csv", np_array, delimiter=","
        )

        self.f_name = csv_file_dir_with_special_chars + "\\" + csv_file_name + ".csv"
        return csv_file_dir

    @staticmethod
    def get_path_of_elmfile_data_inside_external_data_directory() -> str:
        """Path inside (i.e. relative to) external data directory of PF where Elmfile data are stored.

        Returns:
            str: path
        """
        return "ElmFiles"

    @staticmethod
    def get_full_path_of_elmfile_data_in_external_data_directory() -> str:
        """Get full path of directory where elmfile data are stored (in PF's exteral working directory)

        Returns:
            str: path
        """
        act_prj = ActiveProjectCached()
        ext_data_dir = act_prj.get_external_data_directory()
        return (
            ext_data_dir
            + "\\"
            + MeasurementFile.get_path_of_elmfile_data_inside_external_data_directory()
        )

    @staticmethod
    def replace_headers_of_csv_file_with_number_of_colums(file_path: str) -> int:
        """Replaces the first row (headers) of a csv file with its number of
        columns. This is needed for import of csv files to PF using ElmFile.
        """
        with (
            open(file_path + ".csv") as read_file,
            open(file_path + ".temp", "w") as write_file,
        ):
            columns_of_first_row = read_file.readline().split(",")
            if columns_of_first_row[-1] == "\n":
                columns_of_first_row = columns_of_first_row[:-1]
            # Minus one because first column is time and should not be counted.
            number_of_columns = len(columns_of_first_row) - 1
            write_file.write(str(number_of_columns) + "\n")
            row = read_file.readline()
            while row:
                write_file.write(row)
                row = read_file.readline()
        replace(file_path + ".temp", file_path + ".csv")
        return number_of_columns

    @staticmethod
    def insert_row_with_number_of_columns_in_csv_file(file_path: str) -> int:
        """Gets the number of columns of the first row in a csv file and
        inserts a row (first row) with this number in the first column.
        This is needed for ElmFile to read csv files.
        """
        with (
            open(file_path + ".csv") as read_file,
            open(file_path + ".temp", "w") as write_file,
        ):
            first_row = read_file.readline()
            columns_of_first_row = first_row.split(",")
            if columns_of_first_row[-1] == "\n":
                columns_of_first_row = columns_of_first_row[:-1]
            # Minus one because first column is time and should not be counted.
            number_of_columns = len(columns_of_first_row) - 1
            write_file.write(str(number_of_columns) + "\n")
            write_file.write(first_row)
            row = read_file.readline()
            while row:
                write_file.write(row)
                row = read_file.readline()
        replace(file_path + ".temp", file_path + ".csv")
        return number_of_columns



import sys

sys.path.insert(0, r".\src")
from warnings import warn

import powfacpy


class PFCgmesInterface(powfacpy.PFActiveProject):
    """Interface for CGMES integration in PowerFactory."""

    def __init__(self, app):
        warn(
            f"{self.__class__.__name__} will be deprecated. Please use the class 'CGMES' from 'applications/model_exchange' instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(app)
        self.ARCHIVE_NAME = "cgmes_archive"
        self.ARCHIVE_FOLDER_NAME = "cgmes_archive_folder"
        self.FILE_TO_ARCHIVE_TOOL_NAME = "cgmes_file_to_archive"
        self.ARCHIVE_TO_FILE_TOOL_NAME = "cgmes_archive_to_file"
        self.ARCHIVE_TO_GRID_TOOL_NAME = "cgmes_archive_to_grid"
        self.GRID_TO_ARCHIVE_TOOL_NAME = "cgmes_grid_to_archive"
        self.EXPORTED_ZIP_NAME = "cgmes_profiles"
        self.CGMES_VERSION = "CGMES 3.0.0"
        self._ALL_PROFILES = "eq tp ssh sc sv dy dl gl"

    # Import functions

    def _create_cim_to_grid_tool(self):
        """Returns the CIM to Grid tool object .ComCimtogrid."""
        cim_to_grid_tool = self.create_in_folder(
            self.ARCHIVE_TO_GRID_TOOL_NAME + ".ComCimtogrid",
            self.app.GetActiveStudyCase(),
            use_existing=True,
        )
        return cim_to_grid_tool

    def _convert_file_to_archive(self, file_path: str, name: str):
        """Convert a .zip CGMES file to a PowerFactory .CimArchive object. Returns the .CimArchive object."""
        file_to_archive_tool = self.create_in_folder(
            self.FILE_TO_ARCHIVE_TOOL_NAME + ".ComCimdbimp",
            self.app.GetActiveStudyCase(),
            use_existing=True,
        )
        file_to_archive_tool.iopt_target = 2
        archive = self._create_archive(name="_".join([self.ARCHIVE_NAME, name]))
        file_to_archive_tool.targetPath = archive
        file_to_archive_tool.fileName = file_path
        file_to_archive_tool.Execute()
        return archive

    def _convert_archive_to_grid(self, cim_archive):
        """Convert a PowerFactory .CimArchive to a grid. Returns None."""
        cim_to_grid_tool = self._create_cim_to_grid_tool()
        cim_to_grid_tool.sourcePath = cim_archive
        cim_to_grid_tool.dependencies = None
        cim_to_grid_tool.partial = 0
        # cim_to_grid_tool.pGrid = [grid] # TODO find out if grid has to be given
        cim_to_grid_tool.Execute()
        return None

    def cgmes_import(self, input_path: str):
        """Converts CGMES .zip files to a PowerFactory grid.

        Args:
            input_path (str): Path to the CGMES .zip archive, containing xml profiles.

        Returns:
            None
        """
        archive = self._convert_file_to_archive(input_path, "imported")
        self._convert_archive_to_grid(archive)
        return None

    def update_profiles(self, update_file_path: str, base_archive):
        """Includes new SSH (and DL) profiles into an already imported grid.

        Args:
            update_file_path (str): Path to the CGMES .zip archive, containing profiles for updating (SSH and optionally DL).
            base_archive (PowerFactory.DataObject|str) Base .CimArchive which the already imported grid has been created from.

        Returns:
            None
        """
        base_archive = self._handle_single_pf_object_or_path_input(base_archive)
        update_profiles_archive = self._convert_file_to_archive(
            update_file_path, "imported_profiles_for_update"
        )
        cim_to_grid_tool = self._create_cim_to_grid_tool()
        cim_to_grid_tool.sourcePath = update_profiles_archive
        cim_to_grid_tool.dependencies = base_archive
        cim_to_grid_tool.partial = 1
        profiles_in_update_file = [
            file.loc_name.split("_")[-2]
            for file in update_profiles_archive.GetContents()
        ]
        self._set_profiles_of_cim_tool(cim_to_grid_tool, self._ALL_PROFILES, False)
        self._set_profiles_of_cim_tool(
            cim_to_grid_tool, " ".join(profiles_in_update_file), True
        )
        # cim_to_grid_tool.pGrid = [grid] # TODO find out if grid has to be given
        cim_to_grid_tool.Execute()
        return None

    # Export functions

    def _create_grid_to_cim_tool(self):
        """Returns the Grid to CIM tool object .ComGridtocim."""
        grid_to_cim_tool = self.create_in_folder(
            self.GRID_TO_ARCHIVE_TOOL_NAME + ".ComGridtocim",
            self.app.GetActiveStudyCase(),
            use_existing=True,
        )
        grid_to_cim_tool.cAuthority = ["Authority1"]  # TODO this probably isn't general
        grid_to_cim_tool.cSelected = [
            1
        ]  # TODO this probably isn't general (should all grids be selected? or actually only the first one?) --> get project with >1 grid and get to run properly
        grid_to_cim_tool.version = self.CGMES_VERSION
        return grid_to_cim_tool

    def _get_archive_folder(self):
        """Get or create the folder for PowerFactory .CimArchive objects. Returns the folder."""
        archive_folder = self.create_in_folder(
            self.ARCHIVE_FOLDER_NAME + ".IntPrjfolder",
            self.app.GetActiveProject(),
            use_existing=True,
            overwrite=False,
        )
        archive_folder.iopt_typ = "cim"
        return archive_folder

    def _create_archive(self, name):
        """Create a PowerFactory .CimArchive object. Returns the .CimArchive object."""
        archive_folder = self._get_archive_folder()
        archive = self.create_in_folder(
            name + ".CimArchive", archive_folder, use_existing=False, overwrite=True
        )
        return archive

    def _set_profiles_of_cim_tool(self, cim_tool, profiles: str, state: bool):
        """Sets the profiles of PowerFactory .ComCimtogrid or .ComGridtocim tools listet in profiles to state.
        Args:
            cim_tool (str): .ComCimtogrid or .ComGridtocim tool.
            profiles (str): String of CGMES profiles to be set in the tool, separeted by single spaces (e.g. 'ssh dl').
            state (bool): State to set the selected prfiles to (True=checked, False=unchecked).
        Returns:
            None
        """
        cim_tool = self._handle_single_pf_object_or_path_input(cim_tool)
        for profile in profiles.split(" "):
            cim_tool.SetAttribute("convert" + profile.upper(), state)
        return None

    def _convert_grid_to_archive(self, selected_profiles: str = "all"):
        """Convert a PowerFactory grid to a .CimArchive. Returns the archive.
        Args:
            selected_profiles (str): String with CGMES profiles to be generated, separeted by single spaces (e.g. 'ssh dl', 'tp'). 'all' selects all profiles.
        Returns:
            None
        """
        grid_to_cim_tool = self._create_grid_to_cim_tool()

        if selected_profiles == "all":
            grid_to_cim_tool.partial = 0  # convert all profiles
        else:
            grid_to_cim_tool.partial = 1  # convert selected profiles
            self._set_profiles_of_cim_tool(grid_to_cim_tool, self._ALL_PROFILES, False)
            self._set_profiles_of_cim_tool(grid_to_cim_tool, selected_profiles, True)

        grid_to_cim_tool.iopt_target = 1  # existing archive
        archive = self._create_archive(
            name="_".join(
                [self.ARCHIVE_NAME, "exported", selected_profiles.replace(" ", "_")]
            )
        )
        grid_to_cim_tool.targetPath = archive
        grid_to_cim_tool.Execute()
        return archive

    def _convert_archive_to_file(self, archive, output_path: str, as_zip: bool = True):
        """Convert a PowerFactory .CimArchive object to a file. Returns None.
        Args:
            archive (PowerFactory .CimArchive object): archive to be converted and saved as a file.
            output_path (str): Path to desired output folder (without file name).
            as_zip (bool): save as .zip file if as_zip, else save as .xml files.

        Returns:
            None
        """
        archive = self._handle_single_pf_object_or_path_input(archive)
        cim_export_tool = self.create_in_folder(
            self.ARCHIVE_TO_FILE_TOOL_NAME + ".ComCimdbexp",
            self.app.GetActiveStudyCase(),
            use_existing=True,
        )
        cim_export_tool.targetFolder = [output_path]
        if as_zip:
            cim_export_tool.zipModels = 0
            cim_export_tool.archiveName = [
                self.EXPORTED_ZIP_NAME,
            ]
        else:
            cim_export_tool.zipModels = 2
        cim_export_tool.sourcePath = archive
        cim_export_tool.Execute()
        return None

    def cgmes_export(
        self, output_path: str, selected_profiles: str = "all", as_zip: bool = True
    ):
        """Exports selected CGMES profiles of active grid.
        Args:
            output_path (str): Path to desired output folder (without file name).
            selected_profiles (str): String with CGMES profiles to be exported, separeted by single spaces (e.g. 'ssh dl', 'tp'). 'all' selects all profiles.
            as_zip (bool): save as .zip file if as_zip, else save as .xml files.

        Returns:
            None
        """
        archive = self._convert_grid_to_archive(selected_profiles=selected_profiles)
        self._convert_archive_to_file(archive, output_path, as_zip=as_zip)
        return None



from powfacpy.result_variables import ResVar
from powfacpy.pf_classes.protocols import PFGeneral, ElmZone
from powfacpy.applications.topology import Topology
from powfacpy.base.active_project import ActiveProject
import sys

import pytest

sys.path.insert(0, r".\src")
RMS_BAL = ResVar.RMS_Bal


@pytest.fixture(scope="function")
def create_zone(
    act_prj: ActiveProject, activate_39_bus_new_england_test_project
) -> tuple[ElmZone, list[PFGeneral]]:
    topo = Topology()
    terminals = act_prj.get_calc_relevant_obj(
        "Bus 36.ElmTerm, Bus 23.ElmTerm, Bus 22.ElmTerm, Bus 35.ElmTerm"
    )
    zone = topo.create_zone("South-East", terminals)
    return zone, terminals


def test_create_boundary(create_zone):
    zone, terminals = create_zone
    all_elms_in_zone = zone.GetAll()
    topo = Topology()

    boundary_without_changing = topo.create_boundary_without_changing_initial_zones(
        "South-East-without-changing", terminals
    )
    assert zone.GetAll() == all_elms_in_zone

    boundary_with_changing = topo.create_boundary_using_intermediate_zone(
        "South-East-with-changing", terminals
    )
    assert not (zone.GetAll() == all_elms_in_zone)

    assert (
        boundary_with_changing.GetInterior() == boundary_without_changing.GetInterior()
    )


if __name__ == "__main__":
    pytest.main([r"tests\applications\test_topology.py"])



from __future__ import annotations

from typing import Callable

from powfacpy.base.active_project import ActiveProjectCached
from powfacpy.pf_classes.protocols import ElmTerm, PFGeneral, StaPll

from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.pf_classes.elm.grouping_base import GroupingBase
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal


class Terminal(ElmBase):

    __slots__ = ()

    def __init__(self, obj: ElmTerm) -> None:
        super().__init__(obj)
        self._obj: ElmTerm

    def __new__(cls, *args, **kwargs) -> ElmTerm | Terminal:
        """Implemented only to add type hints for the created instance.

        Returns:
            ElmTerm | Terminal: New instance
        """
        instance = super().__new__(cls)
        return instance

    def add_phase_locked_loop(
        self,
        folder: PFGeneral | None = None,
        monitor_frequency_pu: bool = False,
        monitor_frequency_Hz: bool = False,
    ) -> StaPll:
        act_prj = ActiveProjectCached()
        if not folder:
            parent_grid = act_prj.get_upstream_obj(
                self._obj, lambda x: x.GetClassName() == "ElmNet"
            )
            folder = act_prj.create_in_folder(
                "Phase locked loops.IntFolder",
                folder=parent_grid,
                overwrite=False,
                use_existing=True,
            )
        pll: StaPll = act_prj.create_in_folder(
            "PLL at " + self._obj.loc_name + ".StaPll", folder=folder
        )
        pll.pbusbar = self._obj
        if monitor_frequency_Hz:
            act_prj.add_results_variable(pll, "s:Fmeas")
        if monitor_frequency_pu:
            act_prj.add_results_variable(pll, "s:fmeas")
        return pll



import sys

sys.path.insert(0, r".\src")
import powfacpy
import importlib

importlib.reload(powfacpy)
from os import getcwd

import pytest

from test_active_project_interface import pfp


@pytest.fixture
def pfri(pf_app):
    # Return PFResultsInterface instance and show app
    pfri = powfacpy.PFResultsInterface(pf_app)
    return pfri


def test_export_to_csv(activate_powfacpy_test_project, pfri):
    study_case_1 = pfri.get_unique_obj(
        r"Study Cases\test_results_interface\Export Simulation 1.IntCase"
    )
    study_case_2 = pfri.get_unique_obj(
        r"Study Cases\test_results_interface\Export Simulation 2.IntCase"
    )
    pfdi = powfacpy.PFDynSimInterface(pfri.app)
    elmres_list = []
    terminal_hv_1 = pfri.get_unique_obj(
        r"Network Model\Network Data\test_active_project_interface\Grid\Terminal HV 2"
    )
    for case in [study_case_1, study_case_2]:
        case.Activate()
        pfdi.add_results_variable(terminal_hv_1, "m:u1")
        pfdi.initialize_and_run_sim()
        elmres_list.append(pfri.get_from_study_case("ElmRes"))
    study_case_1.Activate()
    dir = getcwd() + "\\tests\\tests_output"
    pfri.export_to_csv(dir=dir, file_name="test_1")
    pfri.export_to_csv(dir=dir, file_name="test_2", results_obj=elmres_list[0])
    # lists and different elmres
    pfri.export_to_csv(
        dir=dir,
        file_name="test_3",
        list_of_results_objs=elmres_list,
        elements=[terminal_hv_1, terminal_hv_1],
        variables=["m:u1", "m:u1"],
    )
    # lists and same elmres
    pfri.export_to_csv(
        dir=dir,
        file_name="test_4",
        list_of_results_objs=[elmres_list[0], elmres_list[0]],
        elements=[terminal_hv_1, terminal_hv_1],
        variables=["m:u1", "m:u1"],
    )
    # don't format csv, use comres parameter argument
    pfri.export_to_csv(
        dir=dir,
        file_name="test_5",
        comres_parameters={"iopt_rscl": 1, "scl_start": 0.5},
        format_csv_file=False,
    )

    with pytest.raises(powfacpy.exceptions.PFNotActiveError):
        pfri.app.GetActiveStudyCase().Deactivate()
        pfri.export_to_csv(dir=dir, file_name="test_6")

    with pytest.raises(ValueError):
        # specified results_obj in combination with list_of_results_objs
        pfri.export_to_csv(
            dir=dir,
            file_name="test_exception",
            results_obj=elmres_list[0],
            list_of_results_objs=elmres_list,
            elements=[terminal_hv_1, terminal_hv_1],
            variables=["m:u1", "m:u1"],
        )

    with pytest.raises(ValueError):
        # variables specified without elements
        pfri.export_to_csv(
            dir=dir,
            file_name="test_exception",
            list_of_results_objs=elmres_list,
            # elements=[terminal_hv_1, terminal_hv_1],
            variables=["m:u1", "m:u1"],
        )


def test_export_to_pandas(activate_powfacpy_test_project, pfri):
    pfri.get_unique_obj(r"Study Cases\test_results_interface\Study Case").Activate()
    network_element = pfri.get_obj(
        r"Network Model\Network Data\test_results_interface\Grid\General Load HV.ElmLod",
        include_subfolders=True,
    )[0]
    variables = ["m:i1:bus1", "m:u1:bus1"]
    elmres = pfri.add_results_variable(network_element, variables)
    pfdi = powfacpy.PFDynSimInterface(pfri.app)
    pfdi.initialize_sim(param={"p_resvar": elmres})
    pfdi.run_sim()
    nr_of_columns = len(variables)
    df = pfri.export_to_pandas(
        list_of_results_objs=[
            elmres,
        ]
        * len(variables),
        elements=[
            network_element,
        ]
        * len(variables),
        variables=variables,
    )
    assert len(df.columns) == nr_of_columns

    df = pfri.export_to_pandas()


if __name__ == "__main__":
    pytest.main(([r"tests\deprecated\test_results_interface.py"]))



from powfacpy.base.active_project import ActiveProjectCached, ActiveProject
from powfacpy.pf_class_protocols import PFApp


class ApplicationBase:
    """Base class for applications. Allows to create versions with
    - ActiveProjectCached (recommended when one project stays active)
    - or ActiveProject (recommended when the active project may change)
    """

    def __init__(
        self, pf_app: PFApp | None | bool = False, cached: bool = False
    ) -> None:
        if cached:
            self.act_prj: ActiveProjectCached = ActiveProjectCached(pf_app)
        else:
            self.act_prj: ActiveProject = ActiveProject(pf_app)

    @property
    def app(self):
        return self.act_prj.__class__.app



import os
import glob
import importlib
import sys

import pytest
import numpy as np

sys.path.insert(0, r".\src")

import powfacpy
import powfacpy.applications.model_exchange
from powfacpy.applications.model_exchange import CGMES


importlib.reload(powfacpy)


@pytest.fixture
def pfcgmes(pf_app) -> CGMES:
    # Return PFModelExchangeInterface instance and show app
    pfcgmes = CGMES(pf_app)
    return pfcgmes


def _get_loadflow_results(pfcgmes: CGMES):
    loadflow = pfcgmes.act_prj.app.GetFromStudyCase("ComLdf")
    loadflow.Execute()
    terminals = pfcgmes.act_prj.app.GetCalcRelevantObjects("*.ElmTerm")
    voltages = {term.loc_name: term.GetAttribute("m:u1") for term in terminals}
    lines = pfcgmes.act_prj.app.GetCalcRelevantObjects("*.ElmLne")
    active_powers = {line.loc_name: line.GetAttribute("m:P:bus1") for line in lines}
    return voltages, active_powers


def _clear_output_path(output_path):
    _ = [os.remove(file) for file in glob.glob(os.path.join(output_path, "*"))]
    return None


def _get_output_path():
    return os.path.abspath(r".\tests\tests_output\cgmes_export")


def _activate_new_study_case_and_grid(pfcgmes: CGMES):
    pfcgmes.act_prj.app.GetActiveStudyCase().Deactivate()
    new_study_case = pfcgmes.act_prj.create_in_folder(
        "New Study Case.IntCase",
        r"Study Cases\test_model_exchange_interfaces",
        overwrite=True,
    )
    new_study_case.Activate()
    new_grid = pfcgmes.act_prj.create_in_folder(
        "New Grid.ElmNet",
        r"Network Model\Network Data\test_model_exchange_interfaces",
        overwrite=True,
    )
    new_grid.Activate()
    return new_study_case


def _compare_loadflow_results(before: dict, after: dict, THRESHOLD=1e-4):
    for element_name, value_before in before.items():
        before_after_difference = abs(
            (value_before - after[element_name]) / value_before
        )
        assert before_after_difference < THRESHOLD


def _assert_that_steady_state_changed(u1: dict, u2: dict):
    u1 = np.array(list(u1.values()))
    u2 = np.array(list(u2.values()))
    assert np.abs((u1 - u2) / u1).max() > 0.01


def test_cgmes_export(pfcgmes: CGMES, activate_powfacpy_test_project):
    OUTPUT_PATH = os.path.abspath(r".\tests\tests_output\cgmes_export")
    clear_output_path = lambda: [
        os.remove(file) for file in glob.glob(os.path.join(OUTPUT_PATH, "*"))
    ]
    clear_output_path()
    # TODO create assertion

    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="all", as_zip=True)
    clear_output_path()
    # TODO create assertion

    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="ssh", as_zip=False)
    clear_output_path()
    # TODO create assertion

    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="all", as_zip=False)
    clear_output_path()
    # TODO create assertion

    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="ssh eq", as_zip=True)
    clear_output_path()
    # TODO create assertion


def test_cgmes_export_import(pfcgmes: CGMES, activate_powfacpy_test_project):
    OUTPUT_PATH = _get_output_path()
    _clear_output_path(OUTPUT_PATH)

    pfcgmes.act_prj.get_unique_obj(
        r"Study Cases\test_model_exchange_interfaces\Study Case"
    ).Activate()
    voltages_before, active_powers_before = _get_loadflow_results(pfcgmes)

    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="all", as_zip=True)

    _ = _activate_new_study_case_and_grid(pfcgmes)
    pfcgmes.cgmes_import(OUTPUT_PATH + "\\" + pfcgmes.exported_zip_name + ".zip")

    voltages_after, active_powers_after = _get_loadflow_results(pfcgmes)

    _compare_loadflow_results(voltages_before, voltages_after)
    _compare_loadflow_results(active_powers_before, active_powers_after)


def test_cgmes_update_profiles(pfcgmes: CGMES, activate_powfacpy_test_project):
    OUTPUT_PATH = _get_output_path()
    _clear_output_path(OUTPUT_PATH)

    study_case_source = pfcgmes.act_prj.get_unique_obj(
        r"Study Cases\test_model_exchange_interfaces\Study Case"
    )
    study_case_source.Activate()
    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="all", as_zip=True)

    study_case_destination = _activate_new_study_case_and_grid(pfcgmes)
    pfcgmes.cgmes_import(OUTPUT_PATH + "\\" + pfcgmes.exported_zip_name + ".zip")

    voltages_source_before_changes, active_powers_source_before_changes = (
        _get_loadflow_results(pfcgmes)
    )
    study_case_source.Activate()
    for load in pfcgmes.act_prj.get_obj("*.ElmLod", include_subfolders=True):
        load.qlini = 0
    voltages_source, active_powers_source = _get_loadflow_results(pfcgmes)

    _assert_that_steady_state_changed(voltages_source_before_changes, voltages_source)

    pfcgmes.cgmes_export(OUTPUT_PATH, selected_profiles="ssh dl", as_zip=True)
    study_case_destination.Activate()
    base_archive = pfcgmes.act_prj.get_unique_obj(
        r"cgmes_archive_folder\cgmes_archive_imported"
    )
    pfcgmes.update_profiles(
        OUTPUT_PATH + "\\" + pfcgmes.exported_zip_name + ".zip", base_archive
    )
    voltages_destination, active_powers_destination = _get_loadflow_results(pfcgmes)

    for load in pfcgmes.act_prj.get_obj("*.ElmLod", include_subfolders=True):
        assert load.qlini == 0
    _compare_loadflow_results(voltages_source, voltages_destination)
    _compare_loadflow_results(active_powers_source, active_powers_destination)


if __name__ == "__main__":
    pytest.main(([r"tests\applications\test_model_exchange.py"]))



import sys
import json
import importlib

import pytest
from jsondiff import diff

settings_file = open(".\\settings.json")
settings = json.load(settings_file)
sys.path.append(settings["local path to PowerFactory application"])
import powerfactory

sys.path.insert(0, r".\src")
import powfacpy

importlib.reload(powfacpy)


from test_active_project_interface import pfp


@pytest.fixture
def pfdbi(pf_app):
    # Return PFDataBaseInterface instance
    return powfacpy.PFDatabaseInterface(pf_app)


def test_get_and_set_object_attributes(pfdbi, activate_powfacpy_test_project):
    """
    Tests for getting and setting object attributes.
    The methods get_object_attributes and set_object_attributes can best be tested
    together by
      1. getting data of PF objects from the PF database.
      2. modifiy some of the data.
      3. setting the data of the PF objects in the PF database with the output from 2.
      4. again getting data of PF objects from the PF database.
      5. Checking if the output from 2. and 4. are equal. If so, getting and setting
      data works correctly.
    This process is done with various optional arguments (relative_paths, pf_obj_handling_options).
    The json files under 'tests\tests_output\test_get_object_attributes*.json' can help with the interpretation of the operations.
    """
    truncated_paths = ["", r"Network Model\Network Data"]
    pf_obj_handling_options = ["path", "original_pf_obj"]
    for truncated_path in truncated_paths:
        for pf_obj_handling in pf_obj_handling_options:
            objs = pfdbi.get_obj(
                "*",
                parent_folder=r"Network Model\Network Data\test_database_interface\Grid",
            )
            class_attributes = {
                "*": ["loc_name", "chr_name"],
                "ElmVac": ["bus1", "outserv"],
                "ElmTr2": ["typ_id"],
            }
            obj_attr_dict = pfdbi.get_object_attributes(
                objs,
                class_attributes=class_attributes,
                truncated_path=truncated_path,
                pf_obj_handling=pf_obj_handling,
            )
            if pf_obj_handling == "path":
                with open(
                    r"tests\tests_output\test_get_object_attributes.json", "w"
                ) as json_file:
                    json.dump(obj_attr_dict, json_file, indent=5)
            # Modify data
            for obj in obj_attr_dict.keys():
                obj_attr_dict[obj]["chr_name"] += "A"
            if pf_obj_handling == "path":
                with open(
                    r"tests\tests_output\test_get_object_attributes_modified.json", "w"
                ) as json_file:
                    json.dump(obj_attr_dict, json_file, indent=5)
            pfdbi.set_object_attributes(obj_attr_dict, added_path=truncated_path)
            obj_attr_dict_after_reading_and_writing = pfdbi.get_object_attributes(
                objs,
                class_attributes=class_attributes,
                truncated_path=truncated_path,
                pf_obj_handling=pf_obj_handling,
            )
            assert not diff(obj_attr_dict, obj_attr_dict_after_reading_and_writing)


if __name__ == "__main__":
    pytest.main([r"tests\test_database_interface.py"])
    # pytest.main(([r"tests"]))



from typing import Callable

from icecream import ic

from powfacpy.pf_class_protocols import PFApp, PFGeneral, ElmZone, ElmBoundary
from powfacpy.pf_classes.elm.boundary import Boundary
from powfacpy.applications.application_base import ApplicationBase


class Topology(ApplicationBase):

    def __init__(
        self, pf_app: PFApp | None | bool = False, cached: bool = False
    ) -> None:
        super().__init__(pf_app, cached)

    def create_zone(
        self,
        name: str,
        elms: list[PFGeneral],
        color: int = 2,
        parent_folder: str | PFGeneral = None,
        overwrite: bool = True,
    ) -> ElmZone:
        if not parent_folder:
            parent_folder = self.act_prj.zones_folder
        zone: ElmZone = self.act_prj.create_in_folder(
            name + ".ElmZone", parent_folder, overwrite=overwrite
        )

        zone.icolor = color
        for elm in elms:
            elm.cpZone = zone
        return zone

    def create_boundary_using_intermediate_zone(
        self,
        name: str,
        elms: list[PFGeneral],
        exclude_node_elms: Callable | None = None,
        color: int = 2,
        parent_folder: PFGeneral | str | None = None,
        overwrite: bool = True,
    ) -> ElmBoundary:
        if parent_folder:
            parent_folder = self.act_prj._handle_single_pf_object_or_path_input(
                parent_folder
            )
        else:
            parent_folder = self.act_prj.boundaries_folder
        dummy_zone = self.create_zone(name, elms)
        try:
            if overwrite:
                self.act_prj.delete_obj(
                    name, parent_folder=parent_folder, error_if_non_existent=False
                )
            boundary: ElmBoundary = dummy_zone.DefineBoundary(0)
            if parent_folder:
                parent_folder.Move(boundary)
            if exclude_node_elms:
                boundary = Boundary(boundary)
                boundary.exclude_node_elms_by_condition(exclude_node_elms)
                boundary = boundary._obj
            boundary.icolor = color
        finally:
            dummy_zone.Delete()
        return boundary

    def create_boundary_without_changing_initial_zones(
        self,
        name: str,
        elms: list[PFGeneral],
        exclude_node_elms: Callable | None = None,
        color: int = 2,
        parent_folder: PFGeneral | str | None = None,
        overwrite: bool = True,
    ) -> ElmBoundary:
        if parent_folder:
            parent_folder = self.act_prj._handle_single_pf_object_or_path_input(
                parent_folder
            )
        else:
            parent_folder = self.act_prj.boundaries_folder
        variation = self.act_prj.create_variation("dummy")
        dummy_zone: ElmZone = self.create_zone("dummy", elms)
        try:
            if overwrite:
                self.act_prj.delete_obj(
                    name, parent_folder=parent_folder, error_if_non_existent=False
                )
            boundary: ElmBoundary = dummy_zone.DefineBoundary(0)
            if exclude_node_elms:
                boundary = Boundary(boundary)
                boundary.exclude_node_elms_by_condition(exclude_node_elms)
            cubicles = boundary.cubicles
            orientations = boundary.ciorient
        finally:
            variation.Deactivate()
            variation.Delete()
            boundary = self.act_prj.create_in_folder(
                name + ".ElmBoundary", folder=parent_folder
            )
            for cub, orientation in zip(cubicles, orientations):
                boundary.AddCubicle(cub, orientation)
            boundary.icolor = color
            if parent_folder:
                parent_folder.Move(boundary)
        return boundary



"""
Base classes for elm.
"""

from abc import ABC
from powfacpy.pf_classes.protocols import PFGeneral
from powfacpy.base.base import BaseChildStatic


class ElmBase(BaseChildStatic):
    __slots__ = ()

    def __init__(self, obj: PFGeneral) -> None:
        super().__init__(obj)

    def set_out_of_service(self):
        self._obj.outserv = 1

    def set_into_service(self):
        self._obj.outserv = 0


class SinglePortBase(ABC):
    __slots__ = ()

    @property
    def terminal(self):
        return self.bus1

    @property
    def p(self):
        return self._obj.pgini

    @p.setter
    def p(self, p: float):
        self.pgini = p

    def q(self):
        return self.qgini

    @p.setter
    def q(self, q: float):
        self.qgini = q



import powerfactory

app = powerfactory.GetApplication()

app.PrintPlain("This is an embedded test script.")

#### Insert after this line

# This should be replaced

#### Insert before this line



#### Insert powfacpy after this line

# This should be replaced

#### End powfacpy before this line





import sys

import pytest

sys.path.insert(0, r".\src")
from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.base.active_project import ActiveProject

sys.path.insert(0, r".\tests")


def test_elm_base(act_prj: ActiveProject, activate_powfacpy_test_project):
    terminal = act_prj.get_calc_relevant_obj("*.ElmTerm")[0]
    terminal_elmbase = ElmBase(terminal)

    terminal_elmbase.set_out_of_service()
    assert terminal.outserv == 1
    assert terminal_elmbase.outserv == 1

    terminal_elmbase.set_into_service()
    assert terminal.outserv == 0
    assert terminal_elmbase.outserv == 0


if __name__ == "__main__":
    pytest.main([r"tests\pf_classes\elm\test_elmbase.py"])



from __future__ import annotations

"""Custom exceptions for powfacpy.
"""
import sys
sys.path.insert(0, r'.\src')
import powfacpy


class PFInterfaceError(Exception):
  """There should always be a base class (that inherits
  from 'Exception') for all custom errors/exceptions.
  """
  pass

class PFAttributeError(PFInterfaceError):
  """Attempt to access an invalid attribute of a PF object.
  """
  def __init__(self, obj, msg_raised, pf_active_project: powfacpy.PFActiveProject):
    object_str = pf_active_project.get_path_of_object(obj)
    self.message = (f"Unexpected attribute of object '{object_str}': {msg_raised}.")
    super().__init__(self.message)

class PFObjectAttributeTypeError(PFInterfaceError):
  """Unexpected type of a PF object attribute (e.g. 'None' type).
  """
  def __init__(self, obj, msg_raised, pf_active_project):
    object_str = pf_active_project.get_path_of_object(obj)
    self.message = (f"Unexpected attribute of object '{object_str}': {msg_raised}.")
    super().__init__(self.message) 
     
class PFAttributeTypeError(PFInterfaceError):
  """Attempt to set an invalid type for the attribute of a PF object.
  """
  def __init__(self, obj, attr, msg_raised, pf_active_project):
    object_str = pf_active_project.get_path_of_object(obj)
    self.message = (f"The attribute '{attr}' of the object '{object_str}' is of unexpected type: {msg_raised}.")
    super().__init__(self.message)

class PFPathError(PFInterfaceError):
  """Attempt to access invalid path in PF database.
  """
  def __init__(self, non_existing_child, existing_path):
    self.message = f"'{non_existing_child}' does not exist in '{existing_path}'"
    super().__init__(self.message)

class PFPathInputError(PFInterfaceError):
  """Invalid input for a PF path.
  """
  def __init__(self, path):
    self.message = (f"The path '{path}' is invalid or empty. Please don't start " +
    f"the path with '\\'.")
    super().__init__(self.message)

class PFNonExistingObjectError(PFInterfaceError):
  """Attempt to access PF object (optional: with a specific condition) that does not exist.
  """
  def __init__(self, folder, obj, condition=False, include_subfolders=False):
    folder_str = powfacpy.PFStringManipulation.remove_html_tags_from_path(str(folder))
    folder_str = powfacpy.PFStringManipulation.remove_class_names(folder_str)
    if include_subfolders:
      msg_subfolder = " (and its subfolders)"
    else:
      msg_subfolder = ""
    msg = (f"The folder '{folder_str}'{msg_subfolder} does not contain any object or path named '{obj}'")
    if condition:
      msg = msg + " with specified condition"  
    self.message = msg
    super().__init__(self.message)

class PFNotActiveError(PFInterfaceError):
  """Unexpected inactive PF object.
  """
  def __init__(self, obj_str):
    self.message = (f"Please activate {obj_str}.")
    super().__init__(self.message)

class PFNoActiveStudyCaseError(PFInterfaceError):
  """No active study case found.
  """
  def __init__(self):
    self.message = ("No active study case found. Please activate one.")
    super().__init__(self.message)

class PFAttributeNotSetError(PFInterfaceError):
  """Attempt to access an attribute of a powfacpy class instance,
  but it was not specified.
  """ 
  def __init__(self, attribute_description):
    self.message = (f"Attempt to access {attribute_description}, " 
    "but it was not set.")
    super().__init__(self.message)

class PFCaseStudyParameterValueDefinitionError(PFInterfaceError):
  """Number of parameter values is not the same for every parameter. 
  """
  def __init__(self, par_name, values):
    self.message = (f"Incorrect number of values defined for parameter '{par_name}'. "
    f"Only {len(values)} values were defined.")
    super().__init__(self.message)
    
class PFInconsistentParamValueOfDSLModelInCompositeModel(PFInterfaceError):
  """Attempt to create a dictionary with parameter names and values of DSL
  models in a composite model, but the DSL models have different values for
  the same parameter. 
  """
  def __init__(self, param_name, composite_model):
    self.message = (f"The parameter '{param_name}' takes on different values in "
                    f"the DSL models of the composite model '{composite_model}'. " 
                    "Therefore, creating a single dictionary with correct parameter "
                    "values for all DSL models is not possible.")
    super().__init__(self.message)



from __future__ import annotations

from typing import Callable

from powfacpy.base.active_project import ActiveProjectCached
from powfacpy.applications.results import Results
from powfacpy.pf_classes.protocols import ElmBoundary, PFGeneral
from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.pf_classes.elm.grouping_base import GroupingBase
from powfacpy.pf_classes.set.colscheme import DiagramColorScheme
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal


class Boundary(ElmBase, GroupingBase):

    __slots__ = ()

    def __init__(self, obj: ElmBoundary) -> None:
        super().__init__(obj)
        self._obj: ElmBoundary

    def __new__(cls, *args, **kwargs) -> ElmBoundary | Boundary:
        """Implemented only to add type hints for the created instance.

        Returns:
            ElmBoundary | Boundary: New instance
        """
        instance = super().__new__(cls)
        return instance

    def get_all_internal_elms(
        self,
    ) -> list[PFGeneral]:
        return self._obj.GetInterior()

    def exclude_node_elms_by_condition(self, condition: Callable) -> list[PFGeneral]:
        act_prj = ActiveProjectCached()
        excluded_elms = act_prj.get_by_condition(self._obj.GetInterior(), condition)
        for elm in excluded_elms:
            self._obj.AddCubicle(elm.GetCubicle(0), 1)

    def get_average_frequency(
        self,
        simulation_results: pd.DataFrame,
        source: str = "terminals",
        ignored_elms: list[PFGeneral] | None = None,
    ) -> float:
        """Get the average frequency inside the boundary.

        Args:
            simulation_results (pd.DataFrame): Simulation results (exported using the 'Results' interface)
            source (str, optional): Frequency measurement source. Options:
                - 'terminal' (ElmTerm, default)
                - 'pll' (StaPll)
                - 'sm' (ElmSym)
            ignored_elms (list[PFGeneral] | None, optional): sources (e.g. ElmTerm/StaPll objects) that are not included. Defaults to None.

        Returns:
            float: Average frequency
        """

        def remove_ignored(elms: list):
            if ignored_elms:
                return [elm for elm in elms if not elm in ignored_elms]
            else:
                return elms

        pfres = Results(cached=True)
        if source == "pll":
            plls_in_boundary = remove_ignored(self.get_phase_locked_loops())
            return pfres.get_simulation_results_from_dataframe(
                simulation_results, plls_in_boundary, "s:fmeas"
            ).mean(1)
        elif source == "sm":
            synchronous_machines_in_boundary = remove_ignored(
                self.get_internal_elms(lambda x: x.GetClassName() == "ElmSym")
            )
            return pfres.get_simulation_results_from_dataframe(
                simulation_results,
                synchronous_machines_in_boundary,
                RMS_BAL.ElmSym.s_xspeed.value,
            ).mean(1)
        elif source == "terminal":
            terminals_in_boundary = remove_ignored(
                self.get_internal_elms(lambda x: x.GetClassName() == "ElmTerm")
            )
            return pfres.get_simulation_results_from_dataframe(
                simulation_results, terminals_in_boundary, RMS_BAL.ElmTerm.m_fe.value
            ).mean(1)

    @staticmethod
    def show_boundary_interior_regions_in_network_graphic():
        act_prj = ActiveProjectCached()
        setcolscheme = act_prj.get_diagram_color_scheme()
        DiagramColorScheme(setcolscheme).show_boundary_interior_regions()

    @staticmethod
    def get_P_exchange_res_var_rms_bal() -> str:
        return RMS_BAL.ElmZone.c_Pinter.value

    @staticmethod
    def get_Q_exchange_res_var_rms_bal() -> str:
        return RMS_BAL.ElmZone.c_Qinter.value



import sys

import pytest

sys.path.insert(0, r".\src")
import powfacpy
from powfacpy.applications.networks import Networks
import importlib

importlib.reload(powfacpy)


@pytest.fixture
def pfnet(pf_app) -> Networks:
    return Networks(pf_app)


def test_get_vacant_cubicle_of_terminal(
    pfnet: Networks, activate_powfacpy_test_project
):
    study_case = pfnet.act_prj.get_single_obj(
        r"Study Cases\test_network_interface\Study Case"
    )
    study_case.Activate()
    pfnet.get_vacant_cubicle_of_terminal(
        r"Network Model\Network Data\test_plot_interface\Grid 1\Terminal HV 2"
    )


if __name__ == "__main__":
    pytest.main(([r"tests\applications\test_networks.py"]))



import sys

import pytest

sys.path.insert(0, r".\src")
from powfacpy.pf_classes.elm.zone import Zone
from powfacpy.base.active_project import ActiveProject

from applications.test_topology import create_zone
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal


def test_get_all_internal_elms(create_zone):
    zone, terminals = create_zone
    zone = Zone(zone)
    all_internal_elms = zone.get_all_internal_elms()
    assert set(terminals).issubset(all_internal_elms)


if __name__ == "__main__":
    pytest.main([r"tests\pf_classes\elm\test_zone.py"])



from typing import Any, Callable
from itertools import product
from os import getcwd, makedirs
from os.path import join

from icecream import ic

import powfacpy
from powfacpy.applications.application_base import ApplicationBase
from powfacpy.pf_class_protocols import (
    ElmNet,
    IntCase,
    IntScenario,
    ElmRes,
    IntFolder,
    IntPrjfolder,
    IntScheme,
    PFApp,
)


class StudyCases(ApplicationBase):

    def __init__(
        self, pf_app: PFApp | None | bool = False, cached: bool = False
    ) -> None:
        super().__init__(pf_app, cached)

        self.title: str = "Case_Studies"
        "Title of the case studies."

        # Study Cases Definition
        self.active_grids: ElmNet | list[ElmNet | list[ElmNet]] = None
        "Active grids for each study case (can be multiple for each). If only one is given, the same grid is active in every case."
        self.parameter_values: dict[str, list[Any]] = {}
        "Dictionary with parameters names (keys) and lists with parameter values for each case (values)."
        self.parameter_paths: dict[str, str] = {}
        "Dictionary with parameter names (keys) and their path."
        self.hierarchy: list[str] = []
        "Hierarchy of folders (named after the parameters) where case/scenario/variation objects are located."
        self.omitted_combinations: list[dict[str, list]] = []
        "Omitted parameter combinations (in permutation)."
        self.base_study_case: IntCase | str = None
        "Base study case (or its path) which is copied to create the cases."
        self.parent_folder_study_cases: IntFolder | IntPrjfolder = None
        "Parent folder where study cases are created."
        self.parent_folder_scenarios: IntFolder | IntPrjfolder = None
        "Parent folder where scenarios are created."
        self.parent_folder_variations: IntFolder | IntPrjfolder = None
        "Parent folder where variations are created."

        # Options
        self.add_scenario_to_each_case: bool = True
        "If True, a corresponding scenario is created for each case."
        self.add_variation_to_each_case: bool = False
        "If True, a corresponding variation is created for each case."
        self.consecutively_number_case_names: bool = False
        "If True, numbering is added to the case names."
        self.anonymous_parameters: list[str] = []
        "Parameters for which names are not used in folder/case name strings (only the parameter values are used)."
        self.ignore_parameters_that_are_none_in_names: bool = True
        "If True, parameters with value None are ignored in names (of cases etc.)."
        self.overwrite_study_cases: bool = True
        "If True, existing study cases are overwritten (e.g. when calling 'create_case' one more time with the same settings)."
        self.delimiter: str = " "
        "Default delimiter used in parameter-value strings."

        # Autocreated
        self._study_cases: list[IntCase] = []
        "List of created study case objects when calling 'create_cases'."

    @property
    def study_cases(self) -> list[IntCase]:
        """List of created study case objects when calling 'create_cases' (read-only)."""
        return self._study_cases

    def create_cases(
        self, reactivate_initially_activated_study_case: bool = True
    ) -> None:
        """Create study cases.

        Optionally create corresponding scenarios/variations (if add_scenario_to_each_case/add_variation_to_each_case is True).

        Iterates through all cases and creates study cases (and folders
        according to 'hierarchy') using parameter-value strings for the
        study cases (and folder names).
        """
        if reactivate_initially_activated_study_case:
            initially_active_case = self.act_prj.app.GetActiveStudyCase()
        self._study_cases = []
        number_of_cases = len(next(iter(self.parameter_values.values())))
        if self.base_study_case:
            self.base_study_case = self.act_prj._handle_single_pf_object_or_path_input(
                self.base_study_case
            )
        for case_num in range(number_of_cases):
            folder_path = self.get_folder_path(case_num)
            parameter_values_string = self.get_case_params_value_string(
                case_num, omitted_parameters=self.hierarchy
            )
            if self.consecutively_number_case_names:
                parameter_values_string = str(case_num) + " " + parameter_values_string
            self._study_cases.append(
                self._create_study_case(parameter_values_string, folder_path)
            )
            self.activate_grids(case_num)
            if self.add_scenario_to_each_case:
                scen = self._create_scenario(parameter_values_string, folder_path)
            if self.add_variation_to_each_case:
                self._create_variation(parameter_values_string, folder_path)
            self._set_parameters(case_num)
            if self.add_scenario_to_each_case:
                scen.Save()
        if reactivate_initially_activated_study_case and initially_active_case:
            initially_active_case = self.act_prj.app.GetActiveStudyCase()

    def get_folder_path(self, case_num: int) -> str | None:
        """Get folder path (inside parent folder) of a case.

        The path corresponds to parameter-value pairs specified
        in 'self.hierarchy'.

        Args:
          case_num (int): case number

        Returns:
          str | None: path of study case or None if there is no hierarchy
        """
        if self.hierarchy:
            folder_path = ""
            for par_name in self.hierarchy:
                parameter_value = self.get_value_of_parameter_for_case(
                    par_name, case_num
                )
                if (
                    parameter_value is not None
                    or not self.ignore_parameters_that_are_none_in_names
                ):
                    add_to_string = str(parameter_value) + "\\"
                    if not par_name in self.anonymous_parameters:
                        add_to_string = par_name + "_" + add_to_string
                    folder_path += add_to_string
            if folder_path:
                return folder_path[:-1]  # discard last "\\""
        return None

    def get_value_of_parameter_for_case(
        self, par_name: str, case_obj_or_case_num: IntCase | int
    ) -> Any:
        """Get parameter value for a certain case.

        Note that the values in 'parameter_values' can be

        - a list/tuple where each element corresponds to a case number
        - or a single value which is used for all cases

        Args:
            par_name (str): Parameter name

            case_obj_or_case_num (IntCase | int): Either the case number (int) or
            a study case PF object (then the case number/index is derived first)

        Raises:
            powfacpy.PFCaseStudyParameterValueDefinitionError: If a value is not defined for a certain study case

        Returns:
            Any: a parameter value for a certain case.
        """
        case_num = self._handle_case_input(case_obj_or_case_num)
        values = self.parameter_values[par_name]
        if isinstance(values, (list, tuple)):
            try:
                return values[case_num]
            except IndexError:
                raise powfacpy.exceptions.PFCaseStudyParameterValueDefinitionError(
                    par_name, values
                )
        else:
            return values

    def get_value_of_all_parameters_for_case(
        self, case_obj_or_case_num: IntCase | int
    ) -> list[Any]:
        """Get the value of all paameters for a specific study case.

        Args:
            case_obj_or_case_num (IntCase | int): study case

        Returns:
            list[Any]: parameter values
        """
        parameter_values = []
        case_num = self._handle_case_input(case_obj_or_case_num)
        for par_name in self.parameter_values.keys():
            parameter_values.append(
                self.get_value_of_parameter_for_case(par_name, case_num)
            )
        return parameter_values

    def get_case_params_value_string(
        self,
        case_obj_or_case_num: IntCase | int,
        omitted_parameters: list[str] = None,
        delimiter: str = None,
        equals_sign: str = None,
        anonymous_parameters: list[str] = None,
    ) -> str:
        """Get parameter-value string for a case

        Args:
            case_obj_or_case_num (IntCase | int): study case

            omitted_parameters (list[str], optional): parameters will not be considered. Defaults to None.

            delimiter (str, optional): delimiter between parameter value pairs. Defaults to None.

            equals_sign (str, optional): sign between parameter name and value. Defaults to None.

            anonymous_parameters (list[str], optional): only the value of these parameters will be added (not their name). Defaults to None.

        Returns:
            str: Parameters and their values
        """
        case_num = self._handle_case_input(case_obj_or_case_num)
        if not delimiter:
            delimiter = self.delimiter
        if not equals_sign:
            equals_sign = " _ "
        parameter_values_string = ""
        for par_name in self.parameter_values.keys():
            if omitted_parameters is None or par_name not in omitted_parameters:
                par_value = self.get_value_of_parameter_for_case(par_name, case_num)
                if (
                    par_value is not None
                    or not self.ignore_parameters_that_are_none_in_names
                ):
                    add_to_string = str(par_value) + delimiter
                    if anonymous_parameters is None:
                        if par_name not in self.anonymous_parameters:
                            add_to_string = par_name + equals_sign + add_to_string
                    elif par_name not in anonymous_parameters:
                        add_to_string = par_name + equals_sign + add_to_string
                    parameter_values_string += add_to_string
        # discard last delimiter
        return parameter_values_string[: -len(delimiter)]

    def activate_grids(self, case_num: int) -> None:
        """Activate the corresponding grids of a study case.

        If 'self.active_grids' is a list/tuple, the items correspond to
        each study case. If multiple grids are active for a case, list/tuples can be used in the elements in 'active_grids'.
        If 'self.active_grid' is not a list/tuple, then one grid will be used
        for all cases.

        The grids can be PF objects or their paths.
        """
        if isinstance(self.active_grids, (list, tuple)):
            grids = self.active_grids[case_num]
            if not isinstance(grids, (list, tuple)):
                grids = [grids]
            for grid in grids:
                grid = self.act_prj._handle_single_pf_object_or_path_input(grid)
                grid.Activate()
        elif self.active_grids:
            grid = self.act_prj._handle_single_pf_object_or_path_input(
                self.active_grids
            )
            grid.Activate()

    def set_parent_folders_for_cases_scenarios_variations(
        self, folder_directory: str = ""
    ) -> None:
        """Set the parent folder for the cases, scenarios and variations (if the folders don't exist, a new folder is created).

        Args:
            folder_directory (str, optional):
              directory of folders inside study case/scenarios/variations folder of the project. If it is an empty string, the default project folders are used (e.g. app.GetProjectFolder("study"))
        """
        if folder_directory:
            self.parent_folder_study_cases = self.act_prj.create_directory(
                folder_directory, self.act_prj.app.GetProjectFolder("study")
            )
            self.parent_folder_scenarios = self.act_prj.create_directory(
                folder_directory, self.act_prj.app.GetProjectFolder("scen")
            )
            self.parent_folder_variations = self.act_prj.create_directory(
                folder_directory, self.act_prj.app.GetProjectFolder("scheme")
            )
        else:
            self.parent_folder_study_cases = self.act_prj.app.GetProjectFolder("study")
            self.parent_folder_scenarios = self.act_prj.app.GetProjectFolder("scen")
            self.parent_folder_scenarios = self.act_prj.app.GetProjectFolder("scheme")

    def get_study_cases_parent_folder(self) -> IntFolder | IntPrjfolder:
        """Get folder where study cases are created.

        Returns:
          IntFolder | IntPrjfolder: folder
        """
        if not self.parent_folder_study_cases:
            return self.act_prj.app.GetProjectFolder("study")
        else:
            return self.parent_folder_study_cases

    def get_scenarios_parent_folder(self) -> IntFolder | IntPrjfolder:
        """Get folder where scenarios are created.

        Returns:
            IntFolder | IntPrjfolder: folder
        """
        if not self.parent_folder_scenarios:
            return self.act_prj.app.GetProjectFolder("scen")
        else:
            return self.parent_folder_scenarios

    def get_variations_parent_folder(self) -> IntFolder | IntPrjfolder:
        """Get folder where variations are created.

        Returns:
            IntFolder | IntPrjfolder: folder
        """
        if not self.parent_folder_variations:
            return self.act_prj.app.GetProjectFolder("scheme")
        else:
            return self.parent_folder_variations

    def clear_parent_folders(self):
        """Deletes all objects in the folders returned by
        - self.get_study_cases_parent_folder
        - self.get_scenarios_parent_folder
        - self.get_variations_parent_folder
        """
        parent_folders = [
            self.get_study_cases_parent_folder(),
            self.get_scenarios_parent_folder(),
            self.get_variations_parent_folder(),
        ]
        for parent_folder in parent_folders:
            if parent_folder:
                self.act_prj.clear_folder(parent_folder)

    def get_study_cases(
        self,
        conditions: dict[str, Callable] | Callable,
        return_case_numbers: bool = False,
    ) -> list[IntCase] | tuple[list[IntCase], list[int]]:
        """Retrieve study case objects depending on parameter values.

        Example 1:
          get_study_cases({"par1": lambda x: x == 2, "par2": lambda x: x>0})
            This returns the study cases for which 'par1' equals 2 and 'par2' is
            positive.

        Example 2 (lambda function):
          get_study_cases(lambda x: x[0] >= 2 and x[2] == 'A')

        Args:
            conditions (dict[str, Callable] | Callable):
              Either a dictionary with
                keys: parameter names
                values: Callables with boolean return value depending on
                  parameter (key)

              or a single Callable that accepts an iterable containing all
                parameters.
                Example: lambda x: x[0] >= 2 and x[2] == 'A'

                Note that the order of the parameters in x must be the same as the
                order of the keys in self.parameter_values.

            return_case_numbers (bool, optional): If True, the case number are also returned. Defaults to False.

        Raises:
            ValueError: If 'callable' is invalid.

        Returns:
            list[IntCase] | tuple[list[IntCase], list[int]]: study case objects and case numbers (optional)
        """
        cases_objects = []
        if return_case_numbers:
            case_numbers = []
        # Dictionary with conditions for parameters
        if isinstance(conditions, dict):
            for case_num, case_obj in enumerate(self._study_cases):
                conditions_fullfiled = True
                for parameter, condition in conditions.items():
                    if not condition(
                        self.get_value_of_parameter_for_case(parameter, case_num)
                    ):
                        conditions_fullfiled = False
                        break
                if conditions_fullfiled:
                    cases_objects.append(case_obj)
                    if return_case_numbers:
                        case_numbers.append(case_num)
        # Callables (e.g. lambda function for all parameters)
        elif callable(conditions):
            for case_num, case_obj in enumerate(self._study_cases):
                parameter_values = self.get_value_of_all_parameters_for_case(case_num)
                if conditions(parameter_values):
                    cases_objects.append(case_obj)
                    if return_case_numbers:
                        case_numbers.append(case_num)
        else:
            raise ValueError("conditions must be a dictionary or a callable.")
        if not return_case_numbers:
            return cases_objects
        else:
            return cases_objects, case_numbers

    def get_study_cases_from_string(
        self, conditions: str, return_case_numbers: bool = False
    ) -> list[IntCase] | tuple[list[IntCase], list[int]]:
        """This method is another convenient way to get study cases according to conditions. The conditions are a simple lambda function argument string (see example below). This method is more convenient but less safe than 'self.get_study_cases' because the conditions string is evaluated and a lambda function is created from it. Using eval() statements is generally not recommended due to unforeseeable behavior.
        However, for convenience, it is used here.

        Args:
          conditions (str):
            lambda function argument string:
              Example: "p HV load >= 2 and (control 1 == 'A' and control 2 != 'S')"

          return_case_numbers (bool): If True, not only study case objects, but also study case numbers (indexes) are returned as a tuple.

        Returns:
            list[IntCase] | tuple[list[IntCase], list[int]]: study case objects and case numbers (optional)
        """
        # To create a lambda function from the conditions string, the parameter names need to be replaced by proper python variable names (e.g. "p HV load" is not a proper variable name because of the spaces). Therefore, a dict with the mapping  from parameter names to a list is required (e.g. {"p HV load": x[0],..})
        par_name_to_list_mapping = self._get_parameter_name_to_list_mapping()
        # Then the parameter names are replaced in conditions. Note that there could be strings inside conditions which should not be considered: For example, if there is condition "control == 'control A'", then only the parameter name 'control' should be replaced, not the value 'control A' which also contains 'control', i.e x[0] == 'control A'.
        condition_strings_list = powfacpy.PFStringManipulation.replace_outside_or_inside_of_strings_in_a_string(
            conditions, par_name_to_list_mapping
        )
        lambda_fun = "lambda x: " + "".join(condition_strings_list).strip()
        # Going back to example in the docstring, we now have a lambda function string that can be executed: "lambda x: x[0] >= 2 and (x[1] == 'A' and x[2] != 'S')"
        lambda_fun = eval(lambda_fun)
        cases = self.get_study_cases(
            lambda_fun, return_case_numbers=return_case_numbers
        )
        return cases

    def get_study_case_number(self, study_case: IntCase) -> int:
        """Get the number (index) of a study case object."""
        return self._study_cases.index(study_case)

    def apply_permutation(
        self, omitted_combinations: list[dict[str, list]] = None
    ) -> None:
        """Replaces the values in 'parameter_values' with the permutation of
        their unique elements.

        Use this method if you want to create cases of the permutation of all
        parameters. Note that 'parameter_values' is changed irreversibly.

        Args:
            omitted_combinations (list[dict[str, list]], optional): Parameter combinations that are omitted. Defaults to None.

            Example:
              omitted_combinations = [
                {"Par 1": [1, 0], "Par 2": ["R", "T"]},
                {"Par 3": [2], "Par 1": "all", "Par 4": "all"},
              ]
              Note that the 'all' keyword is used which means that all combiations with this parameter are omitted.
        """
        if omitted_combinations:
            self.omitted_combinations = omitted_combinations
        # Make sure values are unique
        for param_name in self.parameter_values.keys():
            self.parameter_values[param_name] = list(
                set(self.parameter_values[param_name])
            )
        # Use 'product' to get iterable that returns permutation
        permutation_iterable = product(*self.parameter_values.values())
        # Clear values
        original_parameter_values = self.parameter_values.copy()
        for param_name in self.parameter_values.keys():
            self.parameter_values[param_name] = []
        # Copy values from iterable
        for values_of_all_parameters_for_case in permutation_iterable:
            if self.omitted_combinations:
                values_of_all_parameters_for_case = self._filter_omitted_combinations(
                    values_of_all_parameters_for_case, original_parameter_values
                )
            if values_of_all_parameters_for_case:
                for param_num, param_name in enumerate(self.parameter_values.keys()):
                    self.parameter_values[param_name].append(
                        values_of_all_parameters_for_case[param_num]
                    )

    def export_results_of_study_cases_to_csv(
        self,
        export_dir: str = None,
        study_cases: list[IntCase] = None,
        case_numbers: list[int] = None,
        results_obj: str = "ElmRes",
        results_variables: list[str] = None,
        format_csv_file=True,
    ) -> list[str]:
        """Export the simulation results (ElmRes) of the study cases to csv files.
        The csv files are named according to the study case number (e.g. case0.csv, case1.csv,..)
        Returns the full paths of the csv files.

        Arguments:
          - export_dir: directory for export (default is working directory)
          - study_cases: study case objects
          - case_numbers: corresponding study case numbers
          - results_obj: string that is used in GetFromStudyCase to get the ElmRes object (e.g.
            'self.act_prj.app.GetFromStudyCase("ElmRes")' )
          - results_variables: if only specific variables should be export (see also
              export_to_csv). By default all variables are exported.
          - format_csv_file: see export_to_csv
        """
        study_cases, case_numbers = self._handle_study_case_objects_case_numbers_input(
            study_cases=study_cases, case_numbers=case_numbers
        )

        if not export_dir:
            export_dir = getcwd() + "\\" + self.title
        makedirs(export_dir, exist_ok=True)

        csv_files_full_paths = []
        pfri = powfacpy.PFResultsInterface(self.act_prj.app)
        for case_num, case in zip(case_numbers, study_cases):
            case.Activate()
            elmres: ElmRes = self.act_prj.get_from_study_case(results_obj)
            case_file_name = "case" + str(case_num)
            pfri.export_to_csv(
                export_dir,
                case_file_name,
                elmres,
                variables=results_variables,
                format_csv_file=format_csv_file,
            )
            csv_files_full_paths.append(join(export_dir, case_file_name + ".csv"))
        return csv_files_full_paths

    def _handle_case_input(self, case_obj_or_case_num: IntCase | int) -> int:
        """Handle PF study case object (IntCase) or case number (int) input and return the case number.

        If the input is a PF object, the corresponding case number is returned,
        else simply the input (integer) is returned.

        Args:
            case_obj_or_case_num (IntCase | int): study case object (IntCase) or case number (int)

        Returns:
            int: case number
        """
        if not isinstance(case_obj_or_case_num, int):
            return self._study_cases.index(case_obj_or_case_num)
        else:
            return case_obj_or_case_num

    def _create_study_case(self, name: str, folder_path: str) -> IntCase:
        """Create a study case in the folder under 'folder_path' and activates the case.

        Args:
            name (str): name of study case object without class (e.g. parameter-value pairs)

            folder_path (str): relativ to 'self.get_study_cases_parent_folder'

        Returns:
            IntCase: Created study case object
        """
        parent_folder_study_case = self.get_study_cases_parent_folder()
        if folder_path:
            parent_folder_study_case = self.act_prj.create_directory(
                folder_path, parent_folder=parent_folder_study_case
            )
        if not self.base_study_case:
            study_case_obj = self.act_prj.create_in_folder(
                name + ".IntCase",
                parent_folder_study_case,
                overwrite=self.overwrite_study_cases,
            )
        else:
            study_case_obj = self.act_prj.copy_single_obj(
                self.base_study_case,
                parent_folder_study_case,
                overwrite=self.overwrite_study_cases,
                new_name=name,
            )
        study_case_obj.Activate()
        self.act_prj.get_from_study_case("SetDesktop")
        return study_case_obj

    def _create_scenario(self, name: str, folder_path: str) -> IntScenario:
        """Create scenario in folder under 'folder_path'.

        Args:
            name (str): name of scenario (e.g. parameter-value pairs)

            folder_path (str): relative to 'self.get_scenarios_parent_folder'

        Returns:
            IntScenario: Created scenario
        """
        parent_folder_scenario = self.get_scenarios_parent_folder()
        if folder_path:
            parent_folder_scenario = self.act_prj.create_directory(
                folder_path, parent_folder=parent_folder_scenario
            )
        scenario_obj: IntScenario = self.act_prj.create_in_folder(
            name + ".IntScenario", parent_folder_scenario
        )
        scenario_obj.Activate()
        scenario_obj.Save()
        return scenario_obj

    def _create_variation(self, name: str, folder_path: str) -> IntScheme:
        """Create variation with the name 'name' in folder under 'folder_path.

        Args:
            name (str): name of variation (e.g. parameter value pairs)
            folder_path (str): relative to 'self.get_variations_parent_folder'

        Returns:
            IntScheme: Created variation
        """
        parent_folder_variation = self.get_variations_parent_folder()
        if folder_path:
            parent_folder_variation = self.act_prj.create_directory(
                folder_path, parent_folder=parent_folder_variation
            )
        variation_obj = self.act_prj.create_in_folder(
            name + ".IntScheme", parent_folder_variation
        )
        activation_time = self.act_prj.app.GetStudyTimeObject().cTime
        variation_obj.NewStage(name, int(activation_time), 1)
        # variation_obj.Activate()
        return variation_obj

    def _set_parameters(self, case_obj_or_case_num: IntCase | int):
        """Set the parameters according to paths specified in 'self.parameter_paths'
        and values specified in 'self.parameter_values'.
        """
        case_num = self._handle_case_input(case_obj_or_case_num)
        for par_name, path in self.parameter_paths.items():
            value = self.get_value_of_parameter_for_case(par_name, case_num)
            if value is not None:
                self.act_prj.set_attr_by_path(path, value)

    def _filter_omitted_combinations(
        self,
        values_of_all_parameters_for_case: list | tuple,
        original_parameter_values: dict[str, Any],
    ) -> list | None:
        """Filter the parameter combinations that should be omitted. Returns the parameters if the combination is not omitted, otherwise returns None.

        The difficulty is the handling of the 'all' case. In this case, the combination with any value of a parameter is omitted, but combinations of the remaining parameters must be allowed.

        Args:
            values_of_all_parameters_for_case (list | tuple): paramert values for specific case

            original_parameter_values (dict[str, Any]): original values BEFORE permutation is applied.

        Returns:
            list | None: parameters if the combination is not omitted, otherwise returns None
        """
        # Create list from tuple because values might be changed/set to 'None'
        values_of_all_parameters_for_case = list(values_of_all_parameters_for_case)
        for omitted_combination_dict in self.omitted_combinations:
            is_omitted_combination = True
            for param_num, param_name in enumerate(self.parameter_values.keys()):
                if param_name in omitted_combination_dict:
                    if omitted_combination_dict[param_name] == "all":
                        # If the parameter is 'all', then only one case should be created. The case where the value of the parameter takes on the value of the first entry in  original_parameter_values[param_name] is selected. This must be the case for all parameters that are set to 'all'
                        if not (
                            original_parameter_values[param_name][0]
                            == values_of_all_parameters_for_case[param_num]
                        ):
                            return False
                        else:
                            is_omitted_combination = False
                            values_of_all_parameters_for_case[param_num] = None
                    elif (
                        not values_of_all_parameters_for_case[param_num]
                        in omitted_combination_dict[param_name]
                    ):
                        is_omitted_combination = False
                        break
            if is_omitted_combination:
                return False
        return values_of_all_parameters_for_case

    def _get_parameter_name_to_list_mapping(self) -> dict[str, str]:
        """Returns a dictionary that maps the parameters to a string with a list including indexing.

        Example:
          If there are two parameters "p HV load" and "control 1",
          the dictionary is
          {"p HV load": "x[0]", "control 1": "x[1]"}
          {"p HV load": "x[0]", "control 1": "x[1]"}
        """
        par_name_to_list_mapping = {}
        for par_num, par_name in enumerate(self.parameter_values.keys()):
            par_name_to_list_mapping[par_name] = "x[" + str(par_num) + "]"
        return par_name_to_list_mapping

    def _handle_study_case_objects_case_numbers_input(
        self, study_cases: list[IntCase] = None, case_numbers: list[int] = None
    ):
        """Handles input for study case objects and case numbers/indexes.
        Always returns BOTH study case objects and case numbers, independent from
        the input.
        Arguments:
          If only study_cases are provided, case_numbers are inferred.
          If only case_numbers are provided, study_cases are inferred.
          If none are provided, all study cases and case_numbers are returned.
        """
        if not study_cases and not case_numbers:
            # Use all study cases
            study_cases = self._study_cases
            case_numbers = range(len(study_cases))
        elif not case_numbers:
            case_numbers = [self.get_study_case_number(case) for case in study_cases]
        elif not study_cases:
            study_cases = [self._study_cases[case_num] for case_num in case_numbers]
        return study_cases, case_numbers
